/**
 * @author Peter Praxmarer
 * matr#:  9755439
 *
 * @author Bernhard Aichinger
 * matr#:  9825670
 *
 * Implements the logical-time mutual exclusion algorithm
 */

import daj.*;
import java.util.Random;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.NoSuchElementException;


public class LogicalTimeME extends Program {

    protected static final char regionREM = 'R';
    protected static final char regionTRY = 'T';
    protected static final char regionCRIT = 'C';
    protected static final char regionEXIT = 'E';


    protected int rank;
    protected int size;

    // state variables according to the automaton
    public char region;
    public int clock;
    public LinkedList bcast;
    public LinkedList[] send;
    public LinkedList[] history;


    public LogicalTimeME (int rank, int size) { 
        this.rank = rank;
        this.size = size;

        // state variables
        region = regionREM;
        clock = 0;
        bcast = new LinkedList();
        send = new LinkedList[size];
        history = new LinkedList[size];

        for (int i = 0; i < size; i++) {
            send[i] = new LinkedList();
            history[i] = new LinkedList();
        }
    }

    
    protected void dumpHistory() {
        System.out.println("-------------------------------------");
        System.out.println("history of " + rank);
        for (int i = 0; i < history.length; i++) {
            System.out.println("node " + i);
            Iterator iter = history[i].iterator();
            while (iter.hasNext())
                System.out.println(((Msg)iter.next()).getText());
        }
        System.out.println("-------------------------------------");
    }


    /*
     * execution of program
     */
    public void main() { 

        GlobalAssertion assertion = new Assertion();
        Random rnd = new Random();

        while (true) {
	    switch(rnd.nextInt(7)) {
            case 0:      
		// Restricted use of try because it doesn't 
		// make sense that subsequent tries occur - 
		// restriction can be removed
		if(region == regionEXIT || region == regionREM)
		    transitionTRY();
                break;
            case 1:      // bcast
                transitionBCAST();
                break;
            case 2:      // receive
                transitionRECEIVE();
                break;
            case 3:      // send
                transitionSEND();
                break;
            case 4:      // crit
		transitionCRIT();
                break;
            case 5:      // exit
		transitionEXIT();
                break;
            case 6:      // rem
		transitionREM();
                break;
            default:
                break;
            }
            assert(assertion);
        }
    }


    protected void transitionTRY() {
        clock++;
        region = regionTRY;
	bcast.addLast(new Msg(Msg.TRY, clock, rank));
    }


    protected void transitionBCAST() {
        try {
	    do {
		Msg msg = (Msg)bcast.getFirst();
		bcast.removeFirst();
		clock++;
		out().send(msg);
	    } while (true);
	} catch (NoSuchElementException e) {
	}
    }


    protected void transitionRECEIVE() {
        int sender = in().select(1);
        while (sender != -1) {
            Msg msg = (Msg)in(sender).receive();
            if (msg != null) {
                int c = msg.getClock();
                               
                clock = (clock > c ? clock : c) + 1;
                history[msg.getProc()].addLast(msg);
                if (msg.getType() == Msg.TRY && msg.getProc() != rank)
                    send[msg.getProc()].addLast(new Msg(Msg.ACK, clock, rank));
	    }
            sender = in().select(1);
        }
    }


    protected void transitionSEND() {
        for (int j = 0; j < out().getSize(); j++) {
            try {
                Msg msg = (Msg)send[j].getFirst();
		clock++;
                send[j].removeFirst();
		out(j).send(msg);
	    } catch (NoSuchElementException e) {
            }
        }
    }

    private Msg getLast(Iterator iter, int type) {
	Msg ret = null;
	while(iter.hasNext()) {
	    Msg temp = (Msg) iter.next();
	    if(temp.getType() == type || type == Msg.ANY)
		if(ret != null && temp.isBigger(ret) || ret==null)
		    ret = temp;
	}
	return ret;
    }

    private Msg getBigger(Iterator iter, int type, Msg m) {
	Msg ret = m;
	while(iter.hasNext()) {
	    Msg temp = (Msg) iter.next();
	    if( (temp.getType() == type || type == Msg.ANY) 
		&& temp.isBigger(ret) ) 
		ret = temp;
	}
	if (ret.equals(m))
	    return null;
	return ret;
    }

    private Msg getSmaller(Iterator iter, int type, Msg m) {
	Msg ret = m;
	while(iter.hasNext()) {
	    Msg temp = (Msg) iter.next();
	    if( (temp.getType() == type || type == Msg.ANY) && 
		temp.isSmaller(m) ) 
		if(temp.isBigger(ret) || ret == m) //get biggest msg, that is smaller than m
		    ret = temp;
	}
	if (ret.equals(m))
	    return null;
	return ret;
    }

    /**
     * just for cleaning up the histories, improves performance
     */
    private void purgeHistoryUpTo(Msg m) {
	for(int i=0; i<history.length; i++) {
	    Iterator iter = history[i].iterator();
	    LinkedList hist = new LinkedList();
	    while(iter.hasNext()) {
		Msg temp = (Msg) iter.next();
		if (!temp.isSmaller(m))
		    hist.addLast(temp);
	    }
	    history[i] = hist;
	}
    }

    
    
    protected void transitionCRIT() {
	StringBuffer log = new StringBuffer();
	if (region == regionTRY) {
	    Msg lastExit = (Msg) getLast(history[rank].iterator(), Msg.EXIT);
	    
	    Msg lastTry = (Msg) getLast(history[rank].iterator(), Msg.TRY);
	    if(lastTry == null)
		return;

	    log.append("rank=" + rank + ", LastTry: " + lastTry + "\n");

	    if(lastExit != null && lastExit.isBigger(lastTry)) //logicalTimeComparison
		return;

	    log.append("rank=" + rank + ", LastExit: " + lastExit+ "\n");

	    boolean conditionHolds = true;
	    for(int i=0; i < history.length && conditionHolds; i++) {
		if(i == rank) continue;
		
		log.append("rank="+ rank + ", i = " + i + "\n");

		Msg trySmallerThanLastTry = 
		    (Msg) getLast(history[i].iterator(), Msg.TRY);
		conditionHolds = conditionHolds && 
		    (trySmallerThanLastTry == null || 
		     trySmallerThanLastTry.isSmaller(lastTry));

		log.append("rank=" + rank + 
			   ", trySmallerThanLastTry: " + 
			   trySmallerThanLastTry + "\n");

		if(trySmallerThanLastTry != null) {
		    Msg lastExitInHist = 
			(Msg) getBigger(history[i].iterator(), 
					Msg.EXIT, 
					trySmallerThanLastTry);
		    log.append("rank=" + rank + 
			       ", LastExitInHist: " + 
			       lastExitInHist + "\n");
		    conditionHolds = conditionHolds && lastExitInHist != null;
		}

		Msg msg = 
		    (Msg) getBigger(history[i].iterator(), 
				    Msg.ANY /*Msg.ACK*/, lastTry);
		log.append("rank=" + rank + ", msg: " + msg + "\n");

		conditionHolds = conditionHolds && msg != null;
	    }

	    log.append("rank=" + rank + 
		       ", conditionHolds: " + conditionHolds + "\n");

	    if(conditionHolds) {
		clock++;
		region = regionCRIT;
		System.out.println("rank=" + rank + 
				   ", Entered CRIT\nlog:\n" + log );
		purgeHistoryUpTo(lastExit);
		interrupt();
	    }
	}
    }

    protected void transitionEXIT() {
      	clock++;
        region = regionEXIT;
	bcast.addLast(new Msg(Msg.EXIT, clock, rank));
    }


    protected void transitionREM() {
	if(region == regionEXIT) {
	    Msg exitMsg = getLast(history[rank].iterator(), Msg.EXIT);
	    if(exitMsg==null)
		return;

	    Msg tryMsg = getBigger(history[rank].iterator(), Msg.TRY, exitMsg);
	    if(tryMsg != null)
		return;
	    
	    clock++;
	    region = regionREM;
	}
    }


    /*
     * reflects the state of a node
     */
    public String getText() {
        String msgString = "";

        msgString += "region: " + region + "\n";
        msgString += "clock: " + clock;

        return msgString;
    }

    protected void dumpState() {
        System.out.println(">> " + rank + " : " + region + " / " + clock);
    }

}
